#include "consumer_impl.h"
#include "consumer_impl.h"

#include <chrono>

#include "asapo/json_parser/json_parser.h"
#include "asapo/io/io_factory.h"
#include "asapo/http_client/http_error.h"
#include "tcp_consumer_client.h"

#include "asapo/asapo_consumer.h"
#include "fabric_consumer_client.h"
#include "rds_response_error.h"
#include "asapo/common/networking.h"
#include "asapo/common/internal/version.h"

namespace asapo {

ConsumerImpl::ConsumerImpl(std::string server_uri,
                           std::string source_path,
                           bool has_filesystem,
                           SourceCredentials source) :
    io__{GenerateDefaultIO()},
    endpoint_{std::move(server_uri)}, source_path_{std::move(source_path)}, has_filesystem_{has_filesystem},
    source_credentials_(std::move(source)) {

    // net_client__ will be lazy initialized

    if (source_credentials_.data_source.empty()) {
        source_credentials_.data_source = SourceCredentials::kDefaultDataSource;
    }
    if (source_credentials_.instance_id.empty()) {
        source_credentials_.instance_id = SourceCredentials::kDefaultInstanceId;
    }
    if (source_credentials_.pipeline_step.empty()) {
        source_credentials_.pipeline_step = SourceCredentials::kDefaultPipelineStep;
    }

    if (source_credentials_.instance_id == SourceCredentials::kDefaultInstanceId) {
        Error err;
        std::string hostname = io__->GetHostName(&err);

        if (err) {
            hostname = "hostnameerror";
        }

        source_credentials_.instance_id = hostname + "_" + std::to_string(io__->GetCurrentPid());
    }
    if (source_credentials_.pipeline_step == SourceCredentials::kDefaultPipelineStep) {
        source_credentials_.pipeline_step = "DefaultStep";
    }

    service_request_ = std::unique_ptr<ServiceRequest> {new ServiceRequest(endpoint_, source_credentials_)};

    request_sender_details_prefix_ =
            source_credentials_.instance_id + "§" +
            source_credentials_.pipeline_step + "§" +
            source_credentials_.beamtime_id + "§" +
            source_credentials_.data_source + "§";

}

void ConsumerImpl::SetTimeout(uint64_t timeout_ms) {
    timeout_ms_ = timeout_ms;
}

void ConsumerImpl::ForceNoRdma() {
    should_try_rdma_first_ = false;
}

Error ConsumerImpl::DisableMonitoring(bool disable) {
    use_new_api_format_ = !disable;
    return nullptr;
}

NetworkConnectionType ConsumerImpl::CurrentConnectionType() const {
    return current_connection_type_;
}

bool ConsumerImpl::SwitchToGetByIdIfPartialData(Error* err,
                                                const std::string& response,
                                                std::string* group_id,
                                                std::string* redirect_uri) {
    if (*err == ConsumerErrorTemplates::kPartialData) {
        auto error_data = static_cast<const PartialErrorData*>((*err)->GetCustomData());
        if (error_data == nullptr) {
            *err = ConsumerErrorTemplates::kInterruptedTransaction.Generate("malformed response");
            (*err)->AddDetails("response",response);
            return false;
        }
        *redirect_uri = std::to_string(error_data->id);
        *group_id = "0";
        return true;
    }
    return false;
}

bool ConsumerImpl::SwitchToGetByIdIfNoData(Error* err,
                                           const std::string& response,
                                           std::string* group_id,
                                           std::string* redirect_uri) {
    if (*err == ConsumerErrorTemplates::kNoData) {
        auto error_data = static_cast<const ConsumerErrorData*>((*err)->GetCustomData());
        if (error_data == nullptr) {
            *err = ConsumerErrorTemplates::kInterruptedTransaction.Generate("malformed response - " + response);
            return false;
        }
        *redirect_uri = std::to_string(error_data->id);
        *group_id = "0";
        return true;
    }
    return false;
}

Error ConsumerImpl::GetRecordFromServer(std::string* response, std::string group_id, std::string stream,
                                        GetMessageServerOperation op,
                                        bool dataset, uint64_t min_size) {

    if (stream.empty()) {
        return ConsumerErrorTemplates::kWrongInput.Generate("empty stream");
    }

    interrupt_flag_ = false;
    std::string request_suffix = OpToUriCmd(op);
    Error no_data_error;
    auto start = std::chrono::steady_clock::now();
    bool timeout_triggered = false;
    while (true) {
        if (interrupt_flag_) {
            return ConsumerErrorTemplates::kInterruptedTransaction.Generate("interrupted by user request");
        }

        auto ri  = service_request_->CreateBrokerApiRequest(stream, group_id, request_suffix);
        if (dataset) {
            ri.extra_params += "&dataset=true";
            ri.extra_params += "&minsize=" + std::to_string(min_size);
        }
        if (request_suffix == "next" && resend_) {
                ri.extra_params = ri.extra_params + "&resend_nacks=true" + "&delay_ms=" +
                                  std::to_string(delay_ms_) + "&resend_attempts=" + std::to_string(resend_attempts_);
        }
        if (op == GetMessageServerOperation::GetNextAvailable) {
            ri.extra_params += "&id_key=time_id";
        } else {
            ri.extra_params += "&id_key=_id";
        }
        RequestOutput output;
        auto err = service_request_->ServiceRequestWithoutTimeout(ServiceName::kBroker, &current_broker_uri_, ri, &output);
        *response = std::move(output.string_output);
        if (err == nullptr) {
            break;
        }

        if (err == ConsumerErrorTemplates::kStreamFinished) {
            return err;
        }

        if (request_suffix == "next") {
            auto save_error = SwitchToGetByIdIfNoData(&err, *response, &group_id, &request_suffix)
                              || SwitchToGetByIdIfPartialData(&err, *response, &group_id, &request_suffix);
            if (err == ConsumerErrorTemplates::kInterruptedTransaction) {
                return err;
            }
            if (save_error) {
                no_data_error = std::move(err);
            }
        }
        auto elapsed_ms = static_cast<uint64_t>(std::chrono::duration_cast<std::chrono::milliseconds>
            (std::chrono::steady_clock::now() - start).count());
        if (elapsed_ms >= timeout_ms_) {
            if (timeout_triggered || timeout_ms_ == 0) {
                return no_data_error ? std::move(no_data_error) : std::move(err);
            }
            timeout_triggered = true; // to give a chance make another one request if the previous one took too long
        }
        std::this_thread::sleep_for(std::chrono::milliseconds(100));
    }
    return nullptr;
}

Error ConsumerImpl::GetNext(std::string group_id, MessageMeta* info, MessageData* data, std::string stream) {
    return GetMessageFromServer(GetMessageServerOperation::GetNext,
                                0,
                                std::move(group_id),
                                std::move(stream),
                                info,
                                data);
}

Error ConsumerImpl::GetNext(std::string group_id, MessageMeta* info, MessageData* data, std::string stream, bool ordered) {
    if ( ordered ) {
      return GetMessageFromServer(GetMessageServerOperation::GetNext,
                                0,
                                std::move(group_id),
                                std::move(stream),
                                info,
                                data);
    } else {
      return GetMessageFromServer(GetMessageServerOperation::GetNextAvailable,
                                0,
                                std::move(group_id),
                                std::move(stream),
                                info,
                                data);
    }
}

Error ConsumerImpl::GetLast(std::string group_id, MessageMeta* info, MessageData* data, std::string stream) {
    return GetMessageFromServer(GetMessageServerOperation::GetLastInGroup,
                                0,
                                std::move(group_id),
                                std::move(stream),
                                info,
                                data);
}


Error ConsumerImpl::GetLast(MessageMeta* info, MessageData* data, std::string stream) {
    return GetMessageFromServer(GetMessageServerOperation::GetLast,
                                0,
                                "0",
                                std::move(stream),
                                info,
                                data);
}

std::string ConsumerImpl::OpToUriCmd(GetMessageServerOperation op) {
    switch (op) {
    case GetMessageServerOperation::GetNext:
        return "next";
    case GetMessageServerOperation::GetNextAvailable:
        return "next";
    case GetMessageServerOperation::GetLastInGroup:
        return "groupedlast";
    case GetMessageServerOperation::GetLast:
        return "last";
    default:
        return "error";
    }
}

Error ConsumerImpl::GetMessageFromServer(GetMessageServerOperation op, uint64_t id, std::string group_id,
                                         std::string stream,
                                         MessageMeta* info,
                                         MessageData* data) {
    if (info == nullptr) {
        return ConsumerErrorTemplates::kWrongInput.Generate();
    }

    Error err;
    std::string response;
    if (op == GetMessageServerOperation::GetID) {
        err = GetRecordFromServerById(id, &response, std::move(group_id), std::move(stream));
    } else {
        err = GetRecordFromServer(&response, std::move(group_id), std::move(stream), op);
    }
    if (err != nullptr) {
        return err;
    }

    if (!info->SetFromJson(response)) {
        return ConsumerErrorTemplates::kInterruptedTransaction.Generate(std::string("malformed response:") + response);
    }

    return GetDataIfNeeded(info, data);
}

Error ConsumerImpl::GetDataFromFile(MessageMeta* info, MessageData* data) {
    interrupt_flag_ = false;
    uint64_t elapsed_ms = 0;
    Error err;
    while (elapsed_ms <= timeout_ms_) {
        if (interrupt_flag_) {
            err = ConsumerErrorTemplates::kInterruptedTransaction.Generate("interrupted by user request");
            break;
        }
        auto start = std::chrono::steady_clock::now();
        *data = io__->GetDataFromFile(info->FullName(source_path_), &info->size, &err);
        if (err == nullptr) {
            break;
        }
        std::this_thread::sleep_for(std::chrono::milliseconds(100));
        elapsed_ms += static_cast<uint64_t>(std::chrono::duration_cast<std::chrono::milliseconds>
                                            (std::chrono::steady_clock::now() - start).count());
    }
    if (err != nullptr) {
        return ConsumerErrorTemplates::kLocalIOError.Generate(std::move(err));
    }
    return nullptr;
}

bool DataCanBeOnDisk(const MessageMeta* info) {
    return info->ingest_mode ==0 || (info->ingest_mode & IngestModeFlags::kStoreInFilesystem);
}

Error ConsumerImpl::RetrieveData(MessageMeta* info, MessageData* data) {
    if (data == nullptr || info == nullptr) {
        return ConsumerErrorTemplates::kWrongInput.Generate("pointers are empty");
    }

    if (DataCanBeInBuffer(info)) {
        if (TryGetDataFromBuffer(info, data) == nullptr) {
            return nullptr;
        } else {
            info->buf_id = 0;
        }
    }

    if (!DataCanBeOnDisk(info)) {
        return ConsumerErrorTemplates::kDataNotInCache.Generate();
    }

    if (has_filesystem_) {
        return GetDataFromFile(info, data);
    }

    return GetDataFromFileTransferService(info, data, false);
}

Error ConsumerImpl::GetDataIfNeeded(MessageMeta* info, MessageData* data) {
    if (data == nullptr) {
        return nullptr;
    }

    return RetrieveData(info, data);
}


bool ConsumerImpl::DataCanBeInBuffer(const MessageMeta* info) {
    return info->buf_id > 0;
}

Error ConsumerImpl::CreateNetClientAndTryToGetFile(const MessageMeta* info, const std::string& request_sender_details, MessageData* data) {
    const std::lock_guard<std::mutex> lock(net_client_mutex__);
    if (net_client__) {
        return nullptr;
    }

    if (should_try_rdma_first_) { // This will check if a rdma connection can be made and will return early if so
        auto fabricClient = std::unique_ptr<NetClient>(new FabricConsumerClient());

        Error error = fabricClient->GetData(info, request_sender_details, data);

        // Check if the error comes from the receiver data server (so a connection was made)
        if (!error || error == RdsResponseErrorTemplates::kNetErrorNoData) {
            net_client__.swap(fabricClient);
            current_connection_type_ = NetworkConnectionType::kFabric;
            return error; // Successfully received data and is now using a fabric client
        }

        // An error occurred!

        if (std::getenv("ASAPO_PRINT_FALLBACK_REASON")) {
            std::cout << "Fallback to TCP because error: " << error << std::endl;
        }

        // Retry with TCP
        should_try_rdma_first_ = false;
    }

    // Create regular tcp client
    net_client__.reset(new TcpConsumerClient());
    current_connection_type_ = NetworkConnectionType::kAsapoTcp;

    return net_client__->GetData(info, request_sender_details, data);
}

Error ConsumerImpl::TryGetDataFromBuffer(const MessageMeta* info, MessageData* data) {
    std::string request_sender_details;
    if (use_new_api_format_) {
        request_sender_details = request_sender_details_prefix_ + info->stream;
    }

    if (!net_client__) {
        return CreateNetClientAndTryToGetFile(info, request_sender_details, data);
    }

    return net_client__->GetData(info, request_sender_details, data);
}

std::string ConsumerImpl::GenerateNewGroupId(Error* err) {
    RequestInfo ri;
    ri.api = "/" + kConsumerProtocol.GetBrokerVersion() + "/creategroup";
    ri.post = true;
    return service_request_->BrokerRequestWithTimeout(ri, err, timeout_ms_);
}

Error ConsumerImpl::FtsSizeRequestWithTimeout(MessageMeta* info) {
    RequestInfo ri = CreateFileTransferRequest(info);
    ri.extra_params += "&sizeonly=true";
    ri.output_mode = OutputDataMode::string;
    RequestOutput response;
    auto err = service_request_->ServiceRequestWithTimeout(ServiceName::kFileTransfer, &current_fts_uri_, ri, &response, timeout_ms_);
    if (err) {
        return err;
    }

    auto parser = JsonStringParser(std::move(response.string_output));
    err = parser.GetUInt64("file_size", &(info->size));
    return err;
}

Error ConsumerImpl::FtsRequestWithTimeout(MessageMeta* info, MessageData* data) {
    RequestInfo ri = CreateFileTransferRequest(info);
    if (use_new_api_format_) {
        ri.extra_params += "&instanceid=" + service_request_->UrlEscape(source_credentials_.instance_id);
        ri.extra_params += "&pipelinestep=" + service_request_->UrlEscape(source_credentials_.pipeline_step);
        ri.extra_params += "&beamtime=" + service_request_->UrlEscape(source_credentials_.beamtime_id);
        ri.extra_params += "&stream=" + service_request_->UrlEscape(info->stream);
        ri.extra_params += "&source=" + service_request_->UrlEscape(info->source);
    }

    RequestOutput response;
    response.data_output_size = info->size;

    auto err = service_request_->ServiceRequestWithTimeout(ServiceName::kFileTransfer, &current_fts_uri_, ri, &response, timeout_ms_);
    if (err) {
        return err;
    }
    *data = std::move(response.data_output);
    return nullptr;
}

RequestInfo ConsumerImpl::CreateFileTransferRequest(const MessageMeta* info) const {
    RequestInfo ri;
    ri.api = "/" + kConsumerProtocol.GetFileTransferServiceVersion() + "/transfer";
    ri.post = true;
    ri.body = "{\"Folder\":\"" + source_path_ + "\",\"FileName\":\"" + info->name + "\"}";
    ri.cookie = "Authorization=Bearer " + folder_token_;
    ri.output_mode = OutputDataMode::array;
    return ri;
}

Error ConsumerImpl::ResetLastReadMarker(std::string group_id, std::string stream) {
    return SetLastReadMarker(group_id, 0, stream);
}

Error ConsumerImpl::SetLastReadMarker(std::string group_id, uint64_t value, std::string stream) {
    RequestInfo ri = service_request_->CreateBrokerApiRequest(std::move(stream), std::move(group_id), "resetcounter");

    ri.extra_params = "&value=" + std::to_string(value);
    ri.post = true;

    Error err;
    service_request_->BrokerRequestWithTimeout(ri, &err, timeout_ms_);
    return err;
}

uint64_t ConsumerImpl::GetCurrentSize(std::string stream, Error* err) {
    auto ri = GetSizeRequestForSingleMessagesStream(stream);
    return GetCurrentCount(ri, err);
}

Error ConsumerImpl::GetById(uint64_t id, MessageMeta* info, MessageData* data, std::string stream) {
    if (id == 0) {
        return ConsumerErrorTemplates::kWrongInput.Generate("id should be positive");
    }
    return GetMessageFromServer(GetMessageServerOperation::GetID, id, "0", stream, info, data);
}

Error ConsumerImpl::GetRecordFromServerById(uint64_t id, std::string* response, std::string group_id,
                                            std::string stream,
                                            bool dataset, uint64_t min_size) {
    if (stream.empty()) {
        return ConsumerErrorTemplates::kWrongInput.Generate("empty stream");
    }

    RequestInfo ri = service_request_->CreateBrokerApiRequest(std::move(stream), std::move(group_id), std::to_string(id));


    if (dataset) {
        ri.extra_params += "&dataset=true";
        ri.extra_params += "&minsize=" + std::to_string(min_size);
    }

    Error err;
    *response = service_request_->BrokerRequestWithTimeout(ri, &err, timeout_ms_);
    return err;
}

std::string ConsumerImpl::GetBeamtimeMeta(Error* err) {
    RequestInfo ri = service_request_->CreateBrokerApiRequest("default", "0", "meta/0");

    return service_request_->BrokerRequestWithTimeout(ri, err, timeout_ms_);
}

std::string ConsumerImpl::GetStreamMeta(const std::string& stream, Error* err) {
    RequestInfo ri = service_request_->CreateBrokerApiRequest(stream, "0", "meta/1");

    return service_request_->BrokerRequestWithTimeout(ri, err, timeout_ms_);
}


DataSet DecodeDatasetFromResponse(std::string response, Error* err) {
    DataSet res;
    if (!res.SetFromJson(std::move(response))) {
        *err = ConsumerErrorTemplates::kInterruptedTransaction.Generate("malformed response");
        (*err)->AddDetails("response",response);
        return {0, 0, MessageMetas{}};
    } else {
        return res;
    }
}

MessageMetas ConsumerImpl::QueryMessages(std::string query, std::string stream, Error* err) {
    if (stream.empty()) {
        *err = ConsumerErrorTemplates::kWrongInput.Generate("empty stream");
        return {};
    }

    RequestInfo ri = service_request_->CreateBrokerApiRequest(std::move(stream), "0", "querymessages");

    ri.post = true;
    ri.body = std::move(query);

    auto response = service_request_->BrokerRequestWithTimeout(ri, err, timeout_ms_);
    if (*err) {
        return MessageMetas{};
    }

    auto dataset = DecodeDatasetFromResponse("{\"_id\":0,\"size\":0, \"messages\":" + response + "}", err);
    return dataset.content;
}

DataSet ConsumerImpl::GetNextDataset(std::string group_id, uint64_t min_size, std::string stream, Error* err) {
    return GetDatasetFromServer(GetMessageServerOperation::GetNext,
                                0,
                                std::move(group_id),
                                std::move(stream),
                                min_size,
                                err);
}

DataSet ConsumerImpl::GetNextDataset(std::string group_id, uint64_t min_size, std::string stream, bool ordered, Error* err) {
    if ( ordered ) {
    return GetDatasetFromServer(GetMessageServerOperation::GetNextAvailable,
                                0,
                                std::move(group_id),
                                std::move(stream),
                                min_size,
                                err);
    } else {
    return GetDatasetFromServer(GetMessageServerOperation::GetNext,
                                0,
                                std::move(group_id),
                                std::move(stream),
                                min_size,
                                err);
    }
}

DataSet ConsumerImpl::GetLastDataset(uint64_t min_size, std::string stream, Error* err) {
    return GetDatasetFromServer(GetMessageServerOperation::GetLast, 0, "0", std::move(stream), min_size, err);
}

DataSet ConsumerImpl::GetLastDataset(std::string group_id, uint64_t min_size, std::string stream, Error* err) {
    return GetDatasetFromServer(GetMessageServerOperation::GetLastInGroup, 0, std::move(group_id), std::move(stream),
                                min_size, err);
}


DataSet ConsumerImpl::GetDatasetFromServer(GetMessageServerOperation op,
                                           uint64_t id,
                                           std::string group_id, std::string stream,
                                           uint64_t min_size,
                                           Error* err) {
    MessageMetas infos;
    std::string response;
    if (op == GetMessageServerOperation::GetID) {
        *err = GetRecordFromServerById(id, &response, std::move(group_id), std::move(stream), true, min_size);
    } else {
        *err = GetRecordFromServer(&response, std::move(group_id), std::move(stream), op, true, min_size);
    }
    if (*err != nullptr && *err != ConsumerErrorTemplates::kPartialData) {
        return {0, 0, MessageMetas{}};
    }
    return DecodeDatasetFromResponse(response, err);
}

DataSet ConsumerImpl::GetDatasetById(uint64_t id, uint64_t min_size, std::string stream, Error* err) {
    if (id == 0) {
        *err = ConsumerErrorTemplates::kWrongInput.Generate("id should be positive");
        return {};
    }
    return GetDatasetFromServer(GetMessageServerOperation::GetID, id, "0", std::move(stream), min_size, err);
}

StreamInfos ParseStreamsFromResponse(std::string response, Error* err) {
    auto parser = JsonStringParser(std::move(response));
    std::vector<std::string> streams_endcoded;
    StreamInfos streams;
    Error parse_err;
    *err = parser.GetArrayRawStrings("streams", &streams_endcoded);
    if (*err) {
        return StreamInfos{};
    }
    for (auto stream_encoded : streams_endcoded) {
        StreamInfo si;
        auto ok = si.SetFromJson(stream_encoded);
        if (!ok) {
            *err = GeneralErrorTemplates::kSimpleError.Generate("cannot parse " + stream_encoded);
            return StreamInfos{};
        }
        streams.emplace_back(si);
    }
    return streams;
}

std::string filterToString(StreamFilter filter) {
    switch (filter) {
    case StreamFilter::kAllStreams:
        return "all";
    case StreamFilter::kFinishedStreams:
        return "finished";
    case StreamFilter::kUnfinishedStreams:
        return "unfinished";
    }
    return "";
}

StreamInfos ConsumerImpl::GetStreamList(std::string from, StreamFilter filter, Error* err) {
    return GetStreamList(from, filter, true, err);
}

StreamInfos ConsumerImpl::GetStreamList(std::string from, StreamFilter filter, bool detailed, Error* err) {
    RequestInfo ri = GetStreamListRequest(from, filter, detailed);

    auto response = service_request_->BrokerRequestWithTimeout(ri, err, timeout_ms_);
    if (*err) {
        return StreamInfos{};
    }
    return ParseStreamsFromResponse(std::move(response), err);
}

SourceList ConsumerImpl::GetSourceList(Error* err) {
    RequestInfo ri = service_request_->CreateBrokerApiRequest("0", "", "sources");
    ri.post = false;

    auto json_string = service_request_->BrokerRequestWithTimeout(ri, err, timeout_ms_);
    if (*err) {
        return SourceList{};
    }
    SourceList list;
    JsonStringParser parser(json_string);
    if ((*err = parser.GetArrayString("sources", &list))) {
        return SourceList{};
    }
    return list;
}

RequestInfo ConsumerImpl::GetStreamListRequest(const std::string& from, const StreamFilter& filter, bool detailed) const {
    RequestInfo ri = service_request_->CreateBrokerApiRequest("0", "", "streams");
    ri.post = false;
    if (!from.empty()) {
        ri.extra_params = "&from=" + service_request_->UrlEscape(from);
    }
    ri.extra_params += "&filter=" + filterToString(filter);
    std::string detailed_str = detailed ? "true" : "false";
    ri.extra_params += "&detailed=" + detailed_str;
    return ri;
}

Error ConsumerImpl::UpdateFolderTokenIfNeeded(bool ignore_existing) {
    if (!folder_token_.empty() && !ignore_existing) {
        return nullptr;
    }
    folder_token_.clear();

    RequestOutput output;
    RequestInfo ri = CreateFolderTokenRequest();
    auto err = service_request_->ProcessRequest(&output, ri, nullptr);
    if (err) {
        return err;
    }
    folder_token_ = std::move(output.string_output);
    return nullptr;
}

RequestInfo ConsumerImpl::CreateFolderTokenRequest() const {
    RequestInfo ri;
    ri.host = endpoint_;
    ri.api = "/asapo-authorizer/" + kConsumerProtocol.GetAuthorizerVersion() + "/folder";
    ri.post = true;
    ri.body =
        "{\"Folder\":\"" + source_path_ + "\",\"BeamtimeId\":\"" + source_credentials_.beamtime_id + "\",\"Token\":\""
        +
        source_credentials_.user_token + "\"}";
    return ri;
}

Error ConsumerImpl::GetDataFromFileTransferService(MessageMeta* info, MessageData* data,
        bool retry_with_new_token) {
    auto err = UpdateFolderTokenIfNeeded(retry_with_new_token);
    if (err) {
        return err;
    }

    if (info->size == 0) {
        err = FtsSizeRequestWithTimeout(info);
        if (err == ConsumerErrorTemplates::kWrongInput
                && !retry_with_new_token) { // token expired? Refresh token and try again.
            return GetDataFromFileTransferService(info, data, true);
        }
        if (err) {
            return err;
        }
    }

    err = FtsRequestWithTimeout(info, data);
    if (err == ConsumerErrorTemplates::kWrongInput
            && !retry_with_new_token) { // token expired? Refresh token and try again.
        return GetDataFromFileTransferService(info, data, true);
    }
    return err;
}

Error ConsumerImpl::Acknowledge(std::string group_id, uint64_t id, std::string stream) {
    if (stream.empty()) {
        return ConsumerErrorTemplates::kWrongInput.Generate("empty stream");
    }
    RequestInfo ri = service_request_->CreateBrokerApiRequest(std::move(stream), std::move(group_id), std::to_string(id));
    ri.post = true;
    ri.body = "{\"Op\":\"ackmessage\"}";

    Error err;
    service_request_->BrokerRequestWithTimeout(ri, &err, timeout_ms_);
    return err;
}

IdList ConsumerImpl::GetUnacknowledgedMessages(std::string group_id,
                                               uint64_t from_id,
                                               uint64_t to_id,
                                               std::string stream,
                                               Error* error) {
    if (stream.empty()) {
        *error = ConsumerErrorTemplates::kWrongInput.Generate("empty stream");
        return {};
    }
    RequestInfo ri = service_request_->CreateBrokerApiRequest(std::move(stream), std::move(group_id), "nacks");
    ri.extra_params = "&from=" + std::to_string(from_id) + "&to=" + std::to_string(to_id);

    auto json_string = service_request_->BrokerRequestWithTimeout(ri, error, timeout_ms_);
    if (*error) {
        return IdList{};
    }

    IdList list;
    JsonStringParser parser(json_string);
    if ((*error = parser.GetArrayUInt64("unacknowledged", &list))) {
        return IdList{};
    }

    return list;
}

uint64_t ConsumerImpl::GetLastAcknowledgedMessage(std::string group_id, std::string stream, Error* error) {
    if (stream.empty()) {
        *error = ConsumerErrorTemplates::kWrongInput.Generate("empty stream");
        return 0;
    }
    RequestInfo ri = service_request_->CreateBrokerApiRequest(std::move(stream), std::move(group_id), "lastack");

    auto json_string = service_request_->BrokerRequestWithTimeout(ri, error, timeout_ms_);
    if (*error) {
        return 0;
    }

    uint64_t id;
    JsonStringParser parser(json_string);
    if ((*error = parser.GetUInt64("lastAckId", &id))) {
        return 0;
    }

    if (id == 0) {
        *error = ConsumerErrorTemplates::kNoData.Generate();
    }
    return id;
}

void ConsumerImpl::SetResendNacs(bool resend, uint64_t delay_ms, uint64_t resend_attempts) {
    resend_ = resend;
    delay_ms_ = delay_ms;
    resend_attempts_ = resend_attempts;
}

Error ConsumerImpl::NegativeAcknowledge(std::string group_id,
                                        uint64_t id,
                                        uint64_t delay_ms,
                                        std::string stream) {
    if (stream.empty()) {
        return ConsumerErrorTemplates::kWrongInput.Generate("empty stream");
    }
    RequestInfo ri = service_request_->CreateBrokerApiRequest(std::move(stream), std::move(group_id), std::to_string(id));
    ri.post = true;
    ri.body = R"({"Op":"negackmessage","Params":{"DelayMs":)" + std::to_string(delay_ms) + "}}";

    Error err;
    service_request_->BrokerRequestWithTimeout(ri, &err, timeout_ms_);
    return err;
}
void ConsumerImpl::InterruptCurrentOperation() {
    interrupt_flag_ = true;
    service_request_->InterruptCurrentOperation();
}

uint64_t ConsumerImpl::GetCurrentDatasetCount(std::string stream, bool include_incomplete, Error* err) {
    RequestInfo ri = GetSizeRequestForDatasetStream(stream, include_incomplete);
    return GetCurrentCount(ri, err);
}

RequestInfo ConsumerImpl::GetSizeRequestForDatasetStream(std::string& stream, bool include_incomplete) const {
    RequestInfo ri = GetSizeRequestForSingleMessagesStream(stream);
    ri.extra_params = std::string("&incomplete=") + (include_incomplete ? "true" : "false");
    return ri;
}

uint64_t ConsumerImpl::GetCurrentCount(const RequestInfo& ri, Error* err) {
    auto responce = service_request_->BrokerRequestWithTimeout(ri, err, timeout_ms_);
    if (*err) {
        return 0;
    }
    return ParseGetCurrentCountResponce(err, responce);
}

uint64_t ConsumerImpl::ParseGetCurrentCountResponce(Error* err, const std::string& responce) const {
    JsonStringParser parser(responce);
    uint64_t size;
    if ((*err = parser.GetUInt64("size", &size)) != nullptr) {
        return 0;
    }
    return size;
}

RequestInfo ConsumerImpl::GetSizeRequestForSingleMessagesStream(std::string& stream) const {
    RequestInfo ri = service_request_->CreateBrokerApiRequest(std::move(stream), "", "size");
    return ri;
}

Error ConsumerImpl::GetVersionInfo(std::string* client_info, std::string* server_info, bool* supported) {
    if (client_info == nullptr && server_info == nullptr && supported == nullptr) {
        return ConsumerErrorTemplates::kWrongInput.Generate("missing parameters");
    }
    if (client_info != nullptr) {
        *client_info =
            "software version: " + std::string(kVersion) + ", consumer protocol: " + kConsumerProtocol.GetVersion();
    }

    if (server_info != nullptr || supported != nullptr) {
        return service_request_->GetServerVersionInfo("consumer", kConsumerProtocol.GetVersion(), server_info, supported);
    }

    return nullptr;
}

RequestInfo ConsumerImpl::GetDeleteStreamRequest(std::string stream, DeleteStreamOptions options) const {
    RequestInfo ri = service_request_->CreateBrokerApiRequest(std::move(stream), "", "delete");
    ri.post = true;
    ri.body = options.Json();
    return ri;
}

Error ConsumerImpl::DeleteStream(std::string stream, DeleteStreamOptions options) {
    auto ri = GetDeleteStreamRequest(std::move(stream), options);
    Error err;
    service_request_->BrokerRequestWithTimeout(ri, &err, timeout_ms_);
    return err;
}

Error ConsumerImpl::SetStreamPersistent(std::string stream) {
    RequestInfo ri = service_request_->CreateBrokerApiRequest(std::move(stream), "", "persist");
    ri.post = true;
    Error err;
    service_request_->BrokerRequestWithTimeout(ri, &err, timeout_ms_);
    return err;
}

}
